import sheraf
from sheraf.attributes import ModelLoader
from sheraf.attributes.base import BaseAttribute


class ModelAttribute(ModelLoader, BaseAttribute):
    """This attribute references another :class:`~sheraf.models.Model`.

    >>> class Horse(sheraf.Model):
    ...     table = "horse"
    ...     name = sheraf.SimpleAttribute()
    ...
    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.SimpleAttribute()
    ...     horse = sheraf.ModelAttribute(Horse)
    ...
    >>> with sheraf.connection(commit=True):
    ...     jolly = Horse.create(name="Jolly Jumper")
    ...     george = Cowboy.create(name="George Abitbol", horse=jolly)
    ...
    ...     george.horse.name
    'Jolly Jumper'

    The referenced model can be dynamically created if its structure is passed through as a dict:

    >>> with sheraf.connection(commit=True):
    ...     peter = Cowboy.create(name="Peter", horse={"name": "Polly Pumper"})
    ...     assert isinstance(peter.horse, Horse)
    ...     peter.horse.name
    'Polly Pumper'

    When the referenced model is deleted, the value of the attribute becomes ``None``.

    >>> with sheraf.connection(commit=True):
    ...     george = Cowboy.read(george.id)
    ...     jolly.delete()
    ...     assert george.horse is None
    """

    def __init__(self, model=None, **kwargs):
        super().__init__(default=None, model=model, **kwargs)

    def deserialize(self, value):
        try:
            return self.model.read(value)
        except (KeyError, sheraf.exceptions.ModelObjectNotFoundException):
            return None

    def serialize(self, value):
        if value is None:
            return None

        elif isinstance(value, sheraf.IndexedModel):
            return value.identifier

        elif isinstance(value, dict):
            return self.model.create(**value).identifier

        else:
            return value

    def update(
        self,
        old_value,
        new_value,
        addition=True,
        edition=True,
        deletion=False,
        replacement=False,
    ):
        if replacement or old_value is None:
            return self.serialize(new_value)

        return old_value.edit(new_value, addition, edition, deletion, replacement)


class InlineModelAttribute(ModelLoader, BaseAttribute):
    """Class for defining attributes whose value is stored in their owner
    object as a dictionary, as opposed to ModelAttributes where the
    corresponding object 'id' is stored instead.

    Use this class (1) to faster access the objects referenced by this
    attribute (2) to encapsulate data for which direct access through
    tables is not preferred.

    :class:`~sheraf.attributes.inline.InlineModelAttribute` should refer to
    :class:`~sheraf.models.inline.InlineModel`.
    """

    def __init__(self, model=None, **kwargs):
        kwargs.setdefault("default", sheraf.types.SmallDict)
        super().__init__(model=model, **kwargs)

    def deserialize(self, value):
        if value is None:
            return None

        return self.model._decorate(value)

    def serialize(self, value):
        if value is None:
            return None

        elif isinstance(value, sheraf.InlineModel):
            return value.mapping

        elif isinstance(value, dict):
            return self.model.create(**value).mapping

        else:
            return self._default_value(value)

    def update(
        self,
        old_value,
        new_value,
        addition=True,
        edition=True,
        deletion=False,
        replacement=False,
    ):
        if replacement or old_value is None:
            return self.serialize(new_value)

        return old_value.edit(new_value, addition, edition, deletion, replacement)



from BTrees.OOBTree import OOBTree


class LargeDict(OOBTree):
    """A Large Dictionnary based on a Object-Object-BTree.

    LargeDicts keys are ordered and thus they can be iterated:

    >>> mydict = sheraf.types.LargeDict({
    ...     "D": "four",
    ...     "C": "three",
    ...     "B": "two",
    ...     "A": "one",
    ... }) # declaration is unordered
    ...
    >>> list(v for v in mydict.values())
    ... # iteration is ordered
    ['one', 'two', 'three', 'four']

    LargeDicts can also be sliced. Slices return a generator over the values:

    >>> list(mydict["B":"D"])
    ['two', 'three', 'four']
    >>> list(mydict[::-1])
    ['four', 'three', 'two', 'one']

    When using integers as dictionnary keys, be careful that slices behave a
    bit different than they would do for a list, as we use **keys** and not
    indexes.

    >>> mydict = sheraf.types.LargeDict({1:"one", 2:"two", 3:"three"})
    >>> mylist = ["one", "two", "three"]
    >>>
    >>> list(mydict[1:2]) # 1 and 2 are keys
    ['one', 'two']
    >>>
    >>> mylist[1:2] # 1 and 2 are indices
    ['two']
    """

    def __getitem__(self, item):
        if not isinstance(item, slice):
            return OOBTree.__getitem__(self, item)

        # Performance shortcuts
        if item.start is None and item.stop is None:
            if item.step == -1:
                return reversed(OOBTree.values(self))
            if item.step == 1 or item.step is None:
                return OOBTree.values(self)

        keys = OOBTree.keys(self, min=item.start, max=item.stop)
        if item.step and item.step < 0:
            return (
                OOBTree.__getitem__(self, keys[i - 1])
                for i in range(len(keys), 0, item.step)
            )

        return (
            OOBTree.__getitem__(self, keys[i])
            for i in range(0, len(keys), item.step or 1)
        )

    def iterkeys(self, min=None, max=None, reverse=False):
        """Iter over the keys.

        Performances when `reverse` is `True` may be a bit slower.

        :param min: The first key to search.
        :param max: The last key to search.
        :param reverse: If `True`, key order is backward.
        """
        if not reverse:
            return OOBTree.iterkeys(self, min, max)

        return reversed(OOBTree.keys(self, min, max))



import itertools
import operator
import sys
from collections import OrderedDict

from orderedset import OrderedSet

import sheraf.constants
from sheraf.exceptions import InvalidFilterException, InvalidOrderException

from collections.abc import Iterable, Sized


class QuerySet(object):
    """
    A :class:`~sheraf.queryset.QuerySet` is a collection containing
    :class:`~sheraf.models.Model` instances. Like in a regular :class:`set`,
    objects are unique, but the main difference is that
    :class:`~sheraf.queryset.QuerySet` keeps the insertion order.

    :param iterable: A collection of models. If `iterable` is None, then
        `model` must be set.
    :type iterable: Iterable
    :param model_class: A model class to iterate over. If `model` is None, then
        `iterable` must be set. If both are set, `model_class` is ignored.
    :type reversed_iterable: A Model class
    :param predicate: a callable takes an instance as parameter and return a
        Boolean (if True, the instance will be returned in the iteration). If
        None, everything is returned.
    :type predicate: Predicate
    :param kwargs: A dictionnary containing the values expected from the model
        parameters. If `kwargs` is `{"foo": "bar"}` then the queryset will only
        contains models which attribute `foo` is `"bar"`.

    For the following examples, let us work with a simple **Cowboy** model. For
    the sake of simplicity we use a
    :class:`~sheraf.models.IntOrderedNamedAttributesModel` so the first instance
    created will have id 0, the second will have id 1 and so on...

    >>> class Cowboy(sheraf.IntOrderedNamedAttributesModel):
    ...     table = "queryset_people"
    ...     name = sheraf.SimpleAttribute()
    ...     age = sheraf.SimpleAttribute()
    ...
    >>> with sheraf.connection(commit=True):
    ...     peter = Cowboy.create(name="Peter", age=30)
    ...     steven = Cowboy.create(name="Steven", age=30)
    ...     george = Cowboy.create(name="George Abitbol", age=50)

    :class:`~sheraf.queryset.QuerySet` are mostly created by doing requests on a
    Model with :func:`~sheraf.models.indexation.IndexedModel.all`,
    :func:`~sheraf.models.indexation.IndexedModel.filter` or
    :func:`~sheraf.models.indexation.IndexedModel.order`, but can also be
    initialized with custom data.

    >>> with sheraf.connection():
    ...    a = Cowboy.all() # returns a QuerySet with all Cowboy
    ...    b = QuerySet([peter, steven, george]) # This is an equivalent custom QuerySet
    ...    assert set(a) == set(b)

    :class:`~sheraf.queryset.QuerySet` behave like iterators, and can only be
    consumed once.

    >>> with sheraf.connection():
    ...    everybody = Cowboy.all()
    ...    assert {peter, steven, george} == set(everybody)
    ...    assert [] == everybody

    .. note :: :class:`~sheraf.queryset.QuerySet` can be compared against
        anything iterable, but the comparison will consume the
        :class:`~sheraf.queryset.QuerySet`.

    :class:`~sheraf.queryset.QuerySet` keeps the order of insertion.

    >>> assert QuerySet([peter, steven]) != QuerySet([steven, peter])

    :class:`~sheraf.queryset.QuerySet` supports slicing. Slices returns another
    :class:`~sheraf.queryset.QuerySet`.

    >>> with sheraf.connection():
    ...     assert QuerySet([peter]) == Cowboy.all()[0]
    ...     assert QuerySet([george]) == Cowboy.all()[-1]
    ...     assert QuerySet([peter, steven]) == Cowboy.all()[0:2]
    ...     assert QuerySet([peter, steven, george]) == Cowboy.all()[0:]

    :class:`~sheraf.queryset.QuerySet` supports bitwise operations **and**,
    **or** and **xor**. All those operations consume the internal iterator.

    >>> assert QuerySet([peter]) | QuerySet([steven]) == QuerySet([peter, steven])
    >>> assert QuerySet([peter, steven]) & QuerySet([steven]) == QuerySet([steven])
    >>> assert QuerySet([peter, steven]) ^ QuerySet([steven]) == QuerySet([peter])
    """

    def __init__(self, iterable=None, model_class=None, predicate=None, **kwargs):
        self.filters = OrderedDict(kwargs)
        self._iterable = iterable
        self._iterator = None
        self._predicate = predicate
        self.model = model_class
        self.orders = OrderedDict()

        if iterable is None and model_class is None:
            self._iterable = []

        self.__iter__()

    def __iter__(self):
        return self

    def __repr__(self):
        if self.model:
            return "<QuerySet model={}>".format(self.model.__name__)

        if self._iterable:
            return "<QuerySet iterable={}>".format(self._iterable)

        return "<QuerySet>"

    def _model_has_expected_values(self, model):
        for filter_name, expected_value in self.filters.items():
            if getattr(model, filter_name) != expected_value:
                return False

        return True

    def __next__(self):
        if not self._iterator:
            self._init_iterator()

        while True:
            model = next(self._iterator)
            if self._model_has_expected_values(model) and (
                not self._predicate or self._predicate(model)
            ):
                return model

    def __eq__(self, other):
        if isinstance(other, Iterable):
            return all(
                a == b
                for a, b in itertools.zip_longest(
                    iter(self), iter(other), fillvalue=object()
                )
            )
        return super().__eq__(other)

    def __and__(self, other):
        return QuerySet(OrderedSet(self) & OrderedSet(other))

    def __or__(self, other):
        return QuerySet(OrderedSet(self) | OrderedSet(other))

    def __xor__(self, other):
        return QuerySet(OrderedSet(self) ^ OrderedSet(other))

    def count(self):
        """
        :return: The number of objects in the
            :class:`~sheraf.queryset.QuerySet`, but consumes it.

        >>> with sheraf.connection():
        ...     assert Cowboy.create()
        ...     qs = Cowboy.all()
        ...     assert qs.count() == 1
        ...     assert qs.count() == 0
        """
        return sum(1 for _ in self)

    def __getitem__(self, item):
        self._init_iterator()

        if isinstance(item, slice):
            start, stop, step = item.start, item.stop, item.step
        else:
            start, stop, step = item, item + 1, 1

        if self.model:
            maxid = self.model.count()
        elif isinstance(self._iterable, Sized):
            maxid = len(self._iterable)
        elif (start is None or start >= 0) and (stop is None or stop >= 0):
            maxid = None
        else:
            raise ValueError(
                "When a QuerySet contains an unknown sized object, slicing values must be > 0"
            )

        if maxid:
            if isinstance(item, slice):
                start = item.start
                stop = item.stop
                step = item.step
                if start and start < 0:
                    start %= maxid
                if stop and stop < 0:
                    stop %= maxid
            else:
                start, stop, step = item % maxid, (item % maxid) + 1, 1

        # TODO: Avoid to recreate a QuerySet and avoid itertools.islice
        return QuerySet(itertools.islice(self._iterator, start, stop, step))

    def _init_default_iterator(self, reverse=False):
        if self.model:
            keys = self.model.index_iterkeys(reverse=reverse)
            self._iterator = self.model.read_these(keys)

        elif reverse:
            self._iterator = reversed(self._iterable)

        else:
            self._iterator = iter(self._iterable)

    def _init_iterator(self):
        # The default sort order is by ascending identifier
        if not self.orders:
            self._init_default_iterator()
            return

        # If there is only one sort option, and it is over the primary key
        # we can use iterators instead of sorting the whole collection.
        if (
            self.model
            and len(self.orders) == 1
            and self.model.primary_key() in self.orders
        ):
            self._init_default_iterator(
                self.orders[self.model.primary_key()] == sheraf.constants.DESC
            )
            return

        # Else we need to sort the collection.
        # So we successively sort the list from the less important
        # order to the most important order.
        if self._iterable is None:
            self._iterable = self.model.read_these(self.model.index_iterkeys())

        for attribute, order in reversed(self.orders.items()):
            self._iterable = sorted(
                self._iterable,
                key=operator.attrgetter(attribute),
                reverse=(order == sheraf.constants.DESC),
            )

        self._iterator = iter(self._iterable)

    def copy(self):
        """Copies the :class:`~sheraf.queryset.QuerySet` without consuming it.

        >>> with sheraf.connection():
        ...     peter = Cowboy.create(name="Peter")
        ...     steven = Cowboy.create(name="Steven")
        ...     george = Cowboy.create(name="George")
        ...     qall = Cowboy.all()
        ...     qcopy = qall.copy()
        ...
        ...     assert [peter, steven, george] == qall
        ...     # now qall is consumed
        ...
        ...     assert [peter, steven, george] == qcopy
        ...     # now qcopy is consumed
        """

        qs = QuerySet(self._iterable, self.model)
        qs.filters = self.filters.copy()
        qs.orders = self.orders.copy()
        qs._predicate = self._predicate
        return qs

    def delete(self):
        """Delete the objects contained in the queryset.

        Avoids problems when itering on deleted objects.
        """
        identifiers = [(m.__class__, m.identifier) for m in self]
        for klass, identifier in identifiers:
            klass.read(identifier).delete()

    def filter(self, predicate=None, **kwargs):
        """Refine a copy of the current :class:`~sheraf.queryset.QuerySet` with
        further tests.

        :param predicate: filter instance by returning a truthy value.
            If `None` everything is selected.
        :type predicate: callable object
        :param kwargs: A dictionnary containing the values expected from the
            model parameters. If ``kwargs`` is ``{"foo": "bar"}`` then the
            queryset will only contains models which attribute ``foo``
            is ``"bar"``.
        :type kwargs: A dictionary which keys must be valid attributes of the
            model iterated.
        :return: A copy of the current :class:`~sheraf.queryset.QuerySet`
            refined with further tests.
        :return type: :class:`~sheraf.queryset.QuerySet`

        It is possible to chain :func:`~sheraf.queryset.QuerySet.filter` calls:

        >>> with sheraf.connection():
        ...    assert Cowboy.filter(name="George Abitbol", age=50) == \\
        ...           Cowboy.filter(name="George Abitbol").filter(age=50)
        ...
        >>> with sheraf.connection():
        ...    assert Cowboy.filter(lambda person: "Abitbol" in person.name, age=50) == \\
        ...           Cowboy.filter(lambda person: "Abitbol" in person.name).filter(age=50)

        An attribute cannot be filtered twice:

        >>> with sheraf.connection():
        ...    Cowboy.filter(age=30).filter(age=40)
        Traceback (most recent call last):
            ...
        sheraf.exceptions.InvalidFilterException: Some filter parameters appeared twice
        """
        qs = self.copy()

        if self.model:
            for filter_name in kwargs.keys():
                if filter_name not in self.model.attributes:
                    raise sheraf.exceptions.InvalidFilterException(
                        "{} has no attribute {}".format(
                            self.model.__name__, filter_name
                        )
                    )

        common_attributes = set(qs.filters) & set(kwargs)
        invalid_common_attributes = any(
            key for key in common_attributes if qs.filters[key] != kwargs[key]
        )
        if invalid_common_attributes:
            raise InvalidFilterException("Some filter parameters appeared twice")

        qs.filters.update(kwargs)

        if not qs._predicate:
            qs._predicate = predicate

        elif predicate:
            old_predicate = qs._predicate
            qs._predicate = lambda m: old_predicate(m) and predicate(m)

        return qs

    def order(self, *args, **kwargs):
        """Copies the current :class:`~sheraf.queryset.QuerySet` and adds more
        order to it.

        :param args: There can be only one positionnal argument. Choose to iterate
            over ids in an ascending or a descending way.
        :type args: ``sheraf.ASC`` or ``sheraf.DESC``
        :param kwargs: Further parameters will set an order on the matching
            model attributes.
        :type kwargs: A dictionary which keys must be valid attributes of the
            model iterated, and the values must be ``sheraf.ASC`` or
            ``sheraf.DESC``
        :return: A copy of the current :class:`~sheraf.queryset.QuerySet` with
            refined order.
        :return type: :class:`~sheraf.queryset.QuerySet`

        The default order is the ascending model ids.

        >>> with sheraf.connection(commit=True):
        ...     peter = Cowboy.create(name="Peter")
        ...     steven = Cowboy.create(name="Steven")
        ...     george = Cowboy.create(name="George")
        ...
        >>> with sheraf.connection():
        ...     assert [peter, steven, george] == Cowboy.all()
        ...     assert [peter, steven, george] == Cowboy.all().order(sheraf.ASC)
        ...     assert [george, steven, peter] == Cowboy.all().order(sheraf.DESC)
        ...
        ...     assert [george, peter, steven] == Cowboy.all().order(name=sheraf.ASC)
        ...     assert [steven, peter, george] == Cowboy.all().order(name=sheraf.DESC)

        Several order parameters can be passed, either as arguments of the
        function, or by calling :func:`~sheraf.queryset.QuerySet.order` calls.

        .. warning:: Arguments order is only kept since ``Python 3.6``, so
            using multiple `order` parameters will result on undefined order
            for ``Python 3.5-``. If you are using ``Python 3.5-`` and need
            several :class:`~sheraf.queryset.QuerySet` orders, prefer chaining
            `order` method calls.

        >>> with sheraf.connection(): # TODO: not skip # doctest: +SKIP
        ...     assert [george, peter, steven] == Cowboy.all().order(age=sheraf.DESC)
        ...     assert [george, steven, peter] == Cowboy.all().order(age=sheraf.DESC) \\
        ...                                                   .order(name=sheraf.ASC)
        >>> # Only since Python 3.6 # doctest: +SKIP
        ... with sheraf.connection():
        ...     assert [george, steven, peter] == Cowboy.all().order(age=sheraf.DESC, name=sheraf.ASC)

        .. note:: As ids are indexed, sorting on ids only is very performant.
            The less :func:`~sheraf.queryset.QuerySet.order` parameters are
            passed, the better performances will be.
        """
        if not self.model and (args or not kwargs):
            raise InvalidOrderException(
                "QuerySets without models should have explict order arguments"
            )

        if len(args) > 1:
            raise InvalidOrderException(
                "Only one 'order' positionnal parameter is allowed."
            )

        if (sys.version_info.major, sys.version_info.minor) < (3, 6) and len(
            kwargs
        ) > 1:
            raise InvalidOrderException(
                "Only one 'order' named parameter is allowed before python 3.6. Please chain 'order' calls instead."
            )

        qs = self.copy()

        if self.model:
            for attribute, value in kwargs.items():
                if attribute not in self.model.attributes:
                    raise sheraf.exceptions.InvalidOrderException(
                        "{} has no attribute {}".format(self.model.__name__, attribute)
                    )

                if value not in (sheraf.constants.ASC, sheraf.constants.DESC):
                    raise sheraf.exceptions.InvalidOrderException(
                        "Parameter {} has an invalid order value {}".format(
                            attribute, value
                        )
                    )

        if len(args) > 0:
            identifier = args[0]
            if identifier not in (sheraf.constants.ASC, sheraf.constants.DESC):
                raise InvalidOrderException(
                    "Parameter id has an invalid order value {}".format(identifier)
                )

            if self.model.primary_key() in qs.orders:
                raise InvalidOrderException("Id order has been set twice")

            qs.orders[self.model.primary_key()] = identifier

        common_attributes = set(qs.orders) & set(kwargs)
        if common_attributes:
            raise InvalidOrderException("Some order parameters appeared twice")

        qs.orders.update(kwargs)

        return qs

    def get(self):
        """If the :class:`~sheraf.queryset.QuerySet` contains one, and only one
        item, this method returns the item. If the
        :class:`~sheraf.queryset.QuerySet` contains several objects, it raises a
        :class:`~sheraf.exceptions.QuerySetUnpackException`. If the
        :class:`~sheraf.queryset.QuerySet` is empty, it raises a
        :class:`~sheraf.exceptions.EmptyQuerySetUnpackException`.

        >>> with sheraf.connection():
        ...     peter = Cowboy.create(name="Peter")
        ...     steven = Cowboy.create(name="Steven")
        ...     assert peter == Cowboy.filter(name="Peter").get()
        ...     Cowboy.all().get()
        Traceback (most recent call last):
            ...
        sheraf.exceptions.QuerySetUnpackException: Trying to unpack more than 1 value from a QuerySet
        >>> with sheraf.connection():
        ...     Cowboy.filter(age=30).get()
        Traceback (most recent call last):
            ...
        sheraf.exceptions.EmptyQuerySetUnpackException: Trying to unpack an empty QuerySet
        >>> with sheraf.connection():
        ...     Cowboy.filter(name="Unknown cowboy").get()
        Traceback (most recent call last):
            ...
        sheraf.exceptions.EmptyQuerySetUnpackException: Trying to unpack an empty QuerySet
        """
        try:
            element = next(self)
        except StopIteration:
            raise sheraf.exceptions.EmptyQuerySetUnpackException(
                "Trying to unpack an empty QuerySet"
            )

        try:
            next(self)
        except StopIteration:
            return element
        else:
            raise sheraf.exceptions.QuerySetUnpackException(
                "Trying to unpack more than 1 value from a QuerySet"
            )



import persistent
import ZODB.POSException


class CounterMetaclass(type):
    def __new__(cls, name, bases, attrs):
        klass = super().__new__(cls, name, bases, attrs)

        unary_methods = [
            "__abs__",
            "__neg__",
            "__abs__",
            "__invert__",
            "__complex__",
            "__int__",
            "__float__",
            "__str__",
            "__hash__",
        ]

        comparison_methods = [
            "__add__",
            "__and__",
            "__floordiv__",
            "__index__",
            "__invert__",
            "__lshift__",
            "__mod__",
            "__mul__",
            "__matmul__",
            "__or__",
            "__pos__",
            "__pow__",
            "__rshift__",
            "__sub__",
            "__truediv__",
            "__xor__",
            "__div__",
            "__radd__",
            "__rsub__",
            "__rmul__",
            "__rmatmul__",
            "__rtruediv__",
            "__rfloordiv__",
            "__rmod__",
            "__rdivmod__",
            "__rpow__",
            "__rlshift__",
            "__rrshift__",
            "__rand__",
            "__ror__",
            "__rxor__",
            "__rdiv__",
            "__lt__",
            "__le__",
            "__eq__",
            "__ne__",
            "__gt__",
            "__ge__",
        ]

        destructive_methods = [
            "__iadd__",
            "__isub__",
            "__imul__",
            "__imatmul__",
            "__itruediv__",
            "__ifloordiv__",
            "__imod__",
            "__ipow__",
            "__ilshift__",
            "__irshift__",
            "__iand__",
            "__ixor__",
            "__ior__",
        ]

        def comparison_implementation(m):
            def _implementation(self, other):
                if isinstance(other, Counter):
                    return getattr(self.value, m)(other.value)
                return getattr(self.value, m)(other)

            return _implementation

        def unary_implementation(m):
            def _implementation(self):
                return getattr(self.value, m)()

            return _implementation

        def destructive_implementation(m):
            def _implementation(self, other):
                operation_value = other.value if isinstance(other, Counter) else other
                method = m.replace("__i", "__")
                new_value = getattr(self.value, method)(operation_value)

                if new_value == NotImplemented:
                    new_value = getattr(operation_value, method)(self.value)
                if new_value == NotImplemented:
                    raise AttributeError()

                self.value = new_value
                self.nb_editions += 1
                return self

            return _implementation

        for m in unary_methods:
            setattr(klass, m, unary_implementation(m))

        for m in comparison_methods:
            setattr(klass, m, comparison_implementation(m))

        for m in destructive_methods:
            setattr(klass, m, destructive_implementation(m))

        return klass


class Counter(persistent.Persistent, metaclass=CounterMetaclass):
    """Counter is a simple numeric persistent type with conflict proof
    increment and decrement functions."""

    value = 0
    nb_editions = 0

    def __init__(self, value=0):
        if isinstance(value, Counter):
            value = value.value

        self.value = value

    def set(self, v):
        self.value = v
        self.nb_editions += 1

    def increment(self, value):
        self.value += value

    def decrement(self, value):
        self.value -= value

    def _p_resolveConflict(self, old_state, saved_state, new_state):
        old_value, old_nb_editions = old_state
        saved_value, saved_nb_editions = saved_state
        new_value, new_nb_editions = new_state

        if saved_nb_editions != old_nb_editions or new_nb_editions != old_nb_editions:
            raise ZODB.POSException.ConflictError()

        value = saved_value + new_value - old_value
        return (value, old_nb_editions)

    def __getstate__(self):
        return (self.value, self.nb_editions)

    def __setstate__(self, v):
        value, nb_editions = v
        self.value = value
        self.nb_editions = nb_editions

    def __repr__(self):
        return "<Counter value=%s>" % self.value



import BTrees.Length

import sheraf.types.counter
from sheraf.attributes.simples import IntegerAttribute


class CounterAttribute(IntegerAttribute):
    """CounterAttribute is very like
    :class:`~sheraf.attributes.simples.SimpleAttribute` with concurrency-proof
    operations.

    It acts as a drop-in replacement for
    :class:`~sheraf.attributes.simples.SimpleAttribute` with ``increment`` and
    ``decrement`` methods that automatically solve conflicts. It supports all
    mathematical operations, but only solves conflicts using ``increment`` and
    ``decrement``. Every other destructive operation are sensible to conflicts.

    >>> class MyModel(sheraf.Model):
    ...     table = "mymodel"
    ...     # You can just replace SimpleAttribute or IntegerAttribute with CounterAttribute
    ...     counter = sheraf.CounterAttribute() # initialized with 0
    ...
    >>> with sheraf.connection(commit=True):
    ...     m = MyModel.create()
    ...     m.counter.increment(1)
    ...     m.counter.decrement(1)

    Here is how to produce a conflict case, and how CounterAttributes behaves
    to solve it:

    >>> with sheraf.connection(commit=True):
    ...     sheraf.Database.get().nestable = True
    ...     m1 = MyModel.read(m.id)
    ...
    ...     with sheraf.connection(commit=True):
    ...         m2 = MyModel.read(m.id)
    ...         m2.counter.decrement(10)
    ...
    ...     m1.counter.increment(100)
    ...
    >>> with sheraf.connection():
    ...     m3 = MyModel.read(m.id)
    ...     assert 90 == m3.counter

    The conflict resolution understands that a transaction is adding `100` and
    another transaction is substracting `10` to the counter at the same time,
    and finally adds `90`. The formula used is:
    `new_state_a + new_state_b - old_state` where old_state is `0` as it was
    the previous value registered in the database, and `new_state_a` and
    `new_state_b` are the conflicting new values (here `-10` and `100`).

    Using ``+=`` and ``-=`` operators would have raised a conflicts.
    ``increment`` or ``decrement`` must be called explicitely.

    >>> with sheraf.connection(commit=True):
    ...     sheraf.Database.get().nestable = True
    ...     m1 = MyModel.read(m.id)
    ...
    ...     with sheraf.connection(commit=True):
    ...         m2 = MyModel.read(m.id)
    ...         m2.counter -= 10
    ...
    ...     m1.counter += 100
    Traceback (most recent call last):
        ...
    ZODB.POSException.ConflictError: database conflict error ...

    Regular assignments and operations on the counter also raise conflicts
    for automatic conflict resolution.

    >>> with sheraf.connection(commit=True):
    ...     sheraf.Database.get().nestable = True
    ...     m1 = MyModel.read(m.id)
    ...
    ...     with sheraf.connection(commit=True):
    ...         m2 = MyModel.read(m.id)
    ...         m2.counter = 10
    ...
    ...     m1.counter.increment(100)
    Traceback (most recent call last):
        ...
    ZODB.POSException.ConflictError: database conflict error ...
    """

    def __init__(self, default=0, **kwargs):
        """
        :param default: The counter default value. 0 if unset.
        """
        kwargs["lazy"] = False
        super().__init__(
            default=lambda: sheraf.types.counter.Counter(default), **kwargs
        )

    def write(self, parent, value):
        counter = self.read_raw(parent)
        deserialized = self.deserialize(counter)

        if not isinstance(counter, sheraf.types.counter.Counter):
            self.write_raw(parent, deserialized)

        deserialized.set(self.serialize(value))
        return deserialized

    def read(self, parent):
        value = self.read_raw(parent)
        deserialized = self.deserialize(value)

        if not isinstance(value, sheraf.types.counter.Counter):
            self.write_raw(parent, deserialized)

        return deserialized

    def serialize(self, value):
        if isinstance(value, sheraf.types.counter.Counter):
            return value.value

        return value

    def deserialize(self, value):
        if isinstance(value, BTrees.Length.Length):
            return sheraf.types.counter.Counter(value.value)

        if not isinstance(value, sheraf.types.counter.Counter):
            return sheraf.types.counter.Counter(value)

        return value



import itertools
import warnings

import sheraf.exceptions
from sheraf.models.base import BaseModel, BaseModelMetaclass


class BaseIndexedModel(BaseModel):
    """
    This class handles the whole indexation mechanism. The mechanisms
    for reading or iterating over models in the database are handled
    here.
    """

    index_root_default = sheraf.types.SmallDict

    @classmethod
    def primary_key(cls):
        return "id"

    @classmethod
    def all(cls):
        """
        :return: A :class:`~sheraf.queryset.QuerySet` containing all the
            registered models.
        """
        return sheraf.queryset.QuerySet(model_class=cls)

    @classmethod
    def create(cls, *args, **kwargs):
        if hasattr(cls, "make_id"):
            args = list(args)
            identifier = args.pop() if args else kwargs.get(cls.primary_key())
            model = super().create(*args, **kwargs)

            warnings.warn(
                "BaseIndexedModel.make_id is deprecated and wont be supported with sheraf 0.2. "
                "Please use your id attribute 'default' parameter instead.",
                DeprecationWarning,
                stacklevel=3,
            )
            identifier = identifier or model.make_id()
            cls.index_setitem(identifier, model.mapping)
            model.identifier = identifier
        else:
            model = super().create(*args, **kwargs)
            cls.index_setitem(model.identifier, model.mapping)

        return model

    @classmethod
    def read(cls, *args, **kwargs):
        """
        Get a model instance from its identifier. If the model identifier is not valid, a
        :class:`~sheraf.exceptions.ModelObjectNotFoundException` is raised.

        :param *args*: The ``identifier`` of the model. There can be only one positionnal or
                      keyword argument.
        :param *kwargs*: The ``identifier`` of the model. There can be only one positionnal or
                        keyword argument.
        :return: The :class:`~sheraf.models.indexation.BaseIndexedModel` matching the id.

        >>> class MyModel(sheraf.Model):
        ...     table = "my_model"
        ...
        >>> with sheraf.connection():
        ...     m = MyModel.create()
        ...     assert MyModel.read(m.id) == m
        ...     MyModel.read("invalid")
        Traceback (most recent call last):
            ...
        sheraf.exceptions.ModelObjectNotFoundException: Id 'invalid' not found in MyModel
        """

        if len(args) + len(kwargs) != 1:
            raise TypeError(
                "BaseIndexedModel.read takes only one positionnal or named parameter"
            )

        args = list(args)
        identifier = args.pop() if args else kwargs.get(cls.primary_key())

        try:
            mapping = cls.index_getitem(identifier)
            model = cls._decorate(mapping)
            return model
        except (KeyError, TypeError):
            raise sheraf.exceptions.ModelObjectNotFoundException(cls, identifier)

    @classmethod
    def read_these(cls, *args, **kwargs):
        """
        Get model instances from their identifiers. If an instance identifiers does not
        exist, a :class:`~sheraf.exceptions.ModelObjectNotFoundException` is
        raised.

        The function takes only one parameter which key is the index where to
        search, and which values are the index identifier.
        By default the index used is the `id` index.

        :return: A generator over the models matching the keys.

        >>> class MyModel(sheraf.IntIndexedNamedAttributesModel):
        ...     table = "my_model"
        ...
        >>> with sheraf.connection():
        ...     m1 = MyModel.create(id=1)
        ...     m2 = MyModel.create(id=2)
        ...
        ...     assert [m1, m2] == list(MyModel.read_these([m1.id, m2.id]))
        ...     list(MyModel.read_these(["invalid"]))
        Traceback (most recent call last):
            ...
        sheraf.exceptions.ModelObjectNotFoundException: Id 'invalid' not found in MyModel, 'id' index
        """

        if len(args) + len(kwargs) != 1:
            raise TypeError(
                "BaseIndexedModel.read_these takes only one positionnal or named parameter"
            )

        identifiers = args[0] if args else kwargs.get(cls.primary_key())

        return (cls.read(identifier) for identifier in identifiers)

    @classmethod
    def filter(cls, predicate=None, **kwargs):
        """Shortcut for :func:`sheraf.queryset.QuerySet.filter`.

        :return: :class:`sheraf.queryset.QuerySet`
        """
        return sheraf.queryset.QuerySet(model_class=cls).filter(
            predicate=predicate, **kwargs
        )

    @classmethod
    def order(cls, *args, **kwargs):
        """Shortcut for :func:`sheraf.queryset.QuerySet.order`.

        :return: :class:`sheraf.queryset.QuerySet`
        """
        return sheraf.queryset.QuerySet(model_class=cls).order(*args, **kwargs)

    @classmethod
    def get(cls, *args, **kwargs):
        """Shortcut for :func:`sheraf.queryset.QuerySet.filter` and
        :func:`sheraf.queryset.QuerySet.get`. ``Cowboy.get(name="Peter")`` and
        ``Cowboy.filter(name="Peter").get()`` are equivalent.

        :return: The instance of the model if the filter matches exactly one
            instance. Otherwise, it raises a
            :class:`~sheraf.exceptions.QuerySetUnpackException`.

        >>> class Cowboy(sheraf.Model):
        ...     table = "people"
        ...     name = sheraf.SimpleAttribute()
        ...     age = sheraf.SimpleAttribute()
        ...
        >>> with sheraf.connection(commit=True):
        ...     peter = Cowboy.create(name="Peter", age=30)
        ...     steven = Cowboy.create(name="Steven", age=30)
        ...     assert peter == Cowboy.get(name="Peter")
        ...
        >>> with sheraf.connection():
        ...     Cowboy.get()
        Traceback (most recent call last):
            ...
        sheraf.exceptions.QuerySetUnpackException: Trying to unpack more than 1 value from a QuerySet
        >>> with sheraf.connection():
        ...     Cowboy.get(age=30)
        Traceback (most recent call last):
            ...
        sheraf.exceptions.QuerySetUnpackException: Trying to unpack more than 1 value from a QuerySet
        >>> with sheraf.connection():
        ...     Cowboy.get(name="Unknown cowboy")
        Traceback (most recent call last):
            ...
        sheraf.exceptions.EmptyQuerySetUnpackException: Trying to unpack an empty QuerySet
        """
        return cls.filter(*args, **kwargs).get()

    def __repr__(self):
        identifier = (
            self.identifier
            if self.mapping is not None and self.primary_key() in self.mapping
            else None
        )
        return "<{} {}={}>".format(
            self.__class__.__name__, self.primary_key(), identifier
        )

    def __hash__(self):
        return hash(self.identifier)

    def __eq__(self, other):
        return (
            hasattr(self, self.primary_key())
            and hasattr(other, self.primary_key())
            and self.identifier == other.identifier
        )

    @property
    def identifier(self):
        """
        The identifier is the value of the primary_key for the current instance.
        If the primary_key is 'id', then the identifier might be an UUID.
        """
        return getattr(self, self.primary_key())

    def copy(self, **kwargs):
        """
        Copies a model.

        :param \*\*kwargs: Keywords arguments will be passed to
                         :func:`~sheraf.models.BaseModel.create` and thus
                         wont be copied.

        :return: a copy of this instance.
        """

        copy = super().copy(**kwargs)
        if hasattr(self, "make_id"):
            warnings.warn(
                "BaseIndexedModel.make_id is deprecated and wont be supported with sheraf 0.2. "
                "Please use your id attribute 'default' parameter instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            setattr(copy, self.primary_key(), copy.make_id())
        else:
            if self.primary_key():
                self.reset(self.primary_key())

        return copy

    def delete(self):
        """Delete the current model instance.

        >>> class MyModel(sheraf.Model):
        ...     table = "my_model"
        ...
        >>> with sheraf.connection():
        ...    m = MyModel.create()
        ...    assert m == MyModel.read(m.id)
        ...    m.delete()
        ...    m.read(m.id)
        Traceback (most recent call last):
            ...
        sheraf.exceptions.ModelObjectNotFoundException: Id '...' not found in MyModel
        """
        for attr in self.attributes.values():
            attr.delete(self)
        self.index_delete(self.identifier)


class IndexedModelMetaclass(BaseModelMetaclass):
    """
    Contains the mapping of tables (name of models) to their
    corresponding model definitions
    """

    tables = {}

    def __new__(cls, name, bases, attrs):
        klass = super().__new__(cls, name, bases, attrs)

        if "table" in attrs:
            table_name = attrs["table"]
            qualname = attrs["__module__"] + "." + attrs["__qualname__"]

            unique_table_name = attrs.get("unique_table_name", True)
            if (
                unique_table_name
                and table_name in IndexedModelMetaclass.tables
                and name != IndexedModelMetaclass.tables[table_name][0].split(".")[-1]
            ):
                message = "Table named '{table_name}' used twice: {first_class} and {second_class}".format(
                    table_name=table_name,
                    first_class=IndexedModelMetaclass.tables[table_name][0],
                    second_class=qualname,
                )
                raise sheraf.exceptions.SameNameForTableException(message)
            IndexedModelMetaclass.tables[table_name] = (qualname, klass)
        return klass


class IndexedModel(BaseIndexedModel, metaclass=IndexedModelMetaclass):
    """
    Top-level indexed models.

    Those models are stored at the root of the database. They must
    have a **table** parameter defined and an **id** attribute.

    They can have a **database_name** attribute. If it is set, then in a
    default connection context:

    - :func:`~sheraf.models.indexation.IndexedModel.create` will store the\
    new model instances in this database;
    - :func:`~sheraf.models.indexation.IndexedModel.read` and\
    :func:`~sheraf.models.indexation.IndexedModel.all` (etc.) will read in\
    priority in this database, and then in the default database.
    - :func:`~sheraf.models.indexation.IndexedModel.delete` will try to delete\
    the model from this database, and by default in the default database.

    However, if a **database_name** is explicitly passed to
    :func:`sheraf.databases.connection`, then every action will be
    performed on this database, ignoring the model **database_name** attribute.
    """

    database_name = None
    id = sheraf.attributes.simples.SimpleAttribute()

    @staticmethod
    def _current_database_name():
        current_name = sheraf.Database.current_name()
        if not current_name:
            raise sheraf.exceptions.NotConnectedException()
        return current_name

    @staticmethod
    def _table_iterkeys(table, reverse=False):
        try:
            return table.iterkeys(reverse=reverse)
        # TODO: Remove this guard (or just keep it and remove the first lines of this function)
        # on the next compatibility breaking release.
        except TypeError:
            # Some previous stored data may have been OOBTree instead of LargeDict,
            # so the 'reverse' parameter was not available
            keys = table.iterkeys()
            if reverse:
                keys = reversed(list(keys))
            return keys

    @classmethod
    def index_table(cls, database_name=None, setdefault=True):
        index_root = cls.index_root(database_name, setdefault)

        try:
            return index_root[cls.primary_key()]
        except KeyError:
            if not setdefault:
                raise

        return index_root.setdefault(cls.primary_key(), cls.index_table_default())

    @classmethod
    def index_table_default(cls):
        return sheraf.types.LargeDict()

    @classmethod
    def _index_tables(cls, database_name=None):
        tables = []
        for db_name in (database_name, cls.database_name, cls._current_database_name()):
            if not db_name:
                continue

            try:
                tables.append(cls.index_table(db_name, False))
            except KeyError:
                continue

        return tables

    @classmethod
    def index_contains(cls, key):
        return any(key in table for table in cls._index_tables())

    @classmethod
    def index_getitem(cls, key):
        if cls.database_name:
            try:
                return cls.index_table(cls.database_name)[key]
            except KeyError:
                pass
        return cls.index_table(cls._current_database_name())[key]

    @classmethod
    def index_setitem(cls, key, value):
        cls.index_table()[key] = value

    @classmethod
    def index_iterkeys(cls, reverse=False):
        return itertools.chain.from_iterable(
            cls._table_iterkeys(table, reverse=reverse) for table in cls._index_tables()
        )

    @classmethod
    def index_delete(cls, key):
        if cls.database_name:
            try:
                del cls.index_table(cls.database_name)[key]
                return
            except KeyError:
                pass
        del cls.index_table(sheraf.Database.current_name())[key]

    @classmethod
    def index_root(cls, database_name=None, setdefault=True):
        database_name = (
            database_name or cls.database_name or cls._current_database_name()
        )
        root = sheraf.Database.current_connection(database_name).root()

        try:
            return root[cls.table]
        except KeyError:
            if not setdefault:
                raise
            return root.setdefault(cls.table, sheraf.types.SmallDict())

    @classmethod
    def create(cls, *args, **kwargs):
        if cls.primary_key() not in cls.attributes:
            raise sheraf.exceptions.SherafException(
                "{} inherit from IndexedModel but has no id attribute. Cannot create.".format(
                    cls.__name__
                )
            )

        return super().create(*args, **kwargs)

    @classmethod
    def count(cls):
        """
        :return: the number of instances.

        Using this method is faster than using ``len(MyModel.all())``.
        """
        return sum(len(table) for table in cls._index_tables())



import contextlib

import pytest

import sheraf


class Cowboy(sheraf.IntOrderedNamedAttributesModel):
    table = "conftest_people"
    name = sheraf.SimpleAttribute()
    age = sheraf.SimpleAttribute()


@pytest.fixture(autouse=True)
def add_doctest_namespace(doctest_namespace):
    previous_connection = sheraf.connection

    @contextlib.contextmanager
    def magic_connection(*args, **kwargs):
        try:
            sheraf.Database()
        except KeyError:
            pass

        with previous_connection(*args, **kwargs) as connection:
            yield connection

    sheraf.connection = magic_connection
    doctest_namespace["sheraf"] = sheraf
    doctest_namespace["do_amazing_stuff"] = lambda *args, **kwargs: None
    doctest_namespace["Cowboy"] = Cowboy

    yield

    try:
        sheraf.Database.get().close()
    except KeyError:
        pass
    sheraf.connection = previous_connection



import datetime
import uuid
from numbers import Integral

from sheraf.attributes.base import BaseAttribute


class SimpleAttribute(BaseAttribute):
    """Store a primitive data.

    The value can be a :class:`bool`, :class:`str`, :class:`int`,
    :class:`float`.
    """


class TypedAttribute(BaseAttribute):
    """Store a persistent dict of primitive data.

    Keys and values can be :class:`str`, :class:`int`, :class:`float`.
    """

    type = object

    def __init__(self, **kwargs):
        kwargs.setdefault("default", self.type)
        super().__init__(**kwargs)

    def serialize(self, value):
        return self.type(value)


class BooleanAttribute(TypedAttribute):
    """Store a :class:`bool` object."""

    type = bool


class IntegerAttribute(TypedAttribute):
    """Stores an :class:`int` object."""

    type = int


class FloatAttribute(TypedAttribute):
    """Stores a :class:`float` object."""

    type = float


class StringAttribute(TypedAttribute):
    """Stores a :class:`str` object."""

    type = "".__class__


class UUIDAttribute(BaseAttribute):
    """Stores an :class:`uuid.UUID`."""

    def serialize(self, value):
        if value is None:
            return None

        if isinstance(value, Integral):
            return uuid.UUID(int=value).int

        if isinstance(value, (bytes, str, "".__class__)):
            return uuid.UUID(value).int

        return value.int

    def deserialize(self, value):
        if value:
            return uuid.UUID(int=value)
        return None


class StringUUIDAttribute(UUIDAttribute):
    """Stores an :class:`uuid.UUID` but data is handled as a string."""

    def deserialize(self, value):
        uuid = super().deserialize(value)
        if uuid is None:
            return None
        return str(uuid)


class DateTimeAttribute(BaseAttribute):
    """Store a :class:`datetime.datetime` object."""

    def deserialize(self, value):
        if value is None:
            return None

        return datetime.datetime.utcfromtimestamp(value)

    def datetime_to_timestamp(self, date):
        return (date - datetime.datetime(1970, 1, 1)).total_seconds()

    def serialize(self, value):
        if value is None:
            db_value = None

        else:
            value = value.replace(tzinfo=None)
            db_value = self.datetime_to_timestamp(value)

        return db_value


class TimeAttribute(IntegerAttribute):
    """Stores a :class:`datetime.time` object."""

    def __init__(self, default=None, **kwargs):
        super().__init__(default=default, **kwargs)

    def deserialize(self, value):
        if value == -1:
            return None

        seconds = value // 1000000
        microseconds = value % 1000000
        dt = datetime.datetime(1970, 1, 1) + datetime.timedelta(
            seconds=seconds, microseconds=microseconds
        )
        return dt.time()

    def serialize(self, value):
        if value is None:
            return -1

        dt = datetime.datetime(
            1970, 1, 1, value.hour, value.minute, value.second, value.microsecond
        )
        epoch = datetime.datetime(1970, 1, 1)
        delta = dt - epoch
        intvalue = delta.seconds * 1000000 + delta.microseconds
        return intvalue


class DateAttribute(IntegerAttribute):
    """Stores a :class:`datetime.date` object."""

    def __init__(self, default=None, **kwargs):
        super().__init__(default=default, **kwargs)

    def deserialize(self, value):
        if value == -1:
            return None

        return datetime.date(1970, 1, 1) + datetime.timedelta(days=value)

    def serialize(self, value):
        if value is None:
            return -1

        intvalue = (value - datetime.date(1970, 1, 1)).days
        return intvalue



import io
import os

import sheraf
import sheraf.types
from sheraf.attributes.base import BaseAttribute
from sheraf.models.base import BaseModel

FILES_ROOT_DIR = "files/"


def set_files_root_dir(path):
    global FILES_ROOT_DIR
    FILES_ROOT_DIR = path


class FileObjectV1(object):
    def __init__(self, stream="", extension=""):
        # TODO rename stream to content
        self.model = None
        self.attribute_name = None
        self.stream = stream
        self.extension = extension

    def __repr__(self):
        return "<FileObjectV1 model='{}' attribute_name='{}'>".format(
            self.model.__class__.__name__, self.attribute_name
        )

    def __getitem__(self, name):
        return getattr(self, name)

    def __setitem__(self, name, value):
        setattr(self, name, value)

    def __eq__(self, other):
        return (
            isinstance(other, FileObjectV1)
            and self.stream == other.stream
            and self.extension == other.extension
        )

    def keys(self):
        return ["stream", "extension"]

    def associate(self, model, attribute_name):
        self.model = model
        self.attribute_name = attribute_name

    def relative_path(self):
        path = self.absolute_path()
        if not path:
            return None

        return self.relative_path_of(path)

    @staticmethod
    def relative_path_of(path):
        return os.path.relpath(path, FILES_ROOT_DIR)

    def absolute_path(self):
        if not self.model or not self.model.identifier:
            return None

        _dir = self.directory()
        _file_name = "%s.%s" % (self.model.identifier, self.extension)
        return os.path.join(_dir, _file_name)

    def directory(self):
        return self._directory_of(self.model, self.attribute_name)

    @classmethod
    def _directory_of(cls, model, attribute_name):
        return os.path.join(FILES_ROOT_DIR, model.table, attribute_name)

    def exists(self):
        try:
            next(
                self._iter_match_file_paths(
                    self.model, self.attribute_name, ignore_collected_to_remove=True
                )
            )
            return True
        except StopIteration:
            return False

    @classmethod
    def read(cls, model, attribute_name):
        try:
            file_path = next(
                cls._iter_match_file_paths(
                    model, attribute_name, ignore_collected_to_remove=True
                )
            )
            extension = os.path.splitext(file_path)[1].lstrip(".")
            with io.open(file_path, "rb") as f:
                file_instance = cls(extension=extension, stream=f.read())
                file_instance.associate(model, attribute_name)
                return file_instance
        except (StopIteration, OSError):
            pass

    @classmethod
    def _iter_match_file_paths(
        cls, model, attribute_name, ignore_collected_to_remove=False
    ):
        directory = cls._directory_of(model, attribute_name)
        if not os.path.isdir(directory):
            return

        for file_name in os.listdir(directory):
            root = os.path.splitext(file_name)[0]
            if root != model.identifier:
                continue

            path = os.path.join(directory, file_name)
            if ignore_collected_to_remove and path in FilesGarbageCollector.instance():
                continue

            yield path

    def write(self):
        self.delete()
        _path = self.absolute_path()
        _dir = os.path.dirname(_path)
        if not os.path.exists(_dir):
            os.makedirs(_dir)
        with io.open(_path, "wb") as f:
            f.write(self.stream)
            try:
                FilesGarbageCollector.instance().remove(self.relative_path())
            except KeyError:
                pass

    def delete(self):
        for path in self._iter_match_file_paths(self.model, self.attribute_name):
            FilesGarbageCollector.instance().add(path)


class FileObjectV2(FileObjectV1):
    def __init__(self, stream=None, extension=""):
        self.model = None
        self.attribute_name = None
        self.extension = extension
        self._content = stream

    def exists(self):
        if not self._is_new_model():
            return super().exists()

        return os.path.exists(self.absolute_path())

    @property
    def stream(self):
        if self._content is None:
            path = self.absolute_path()
            if path not in FilesGarbageCollector.instance():
                with io.open(self.absolute_path(), "rb") as f:
                    self._content = f.read()

        return self._content

    @stream.setter
    def stream(self, stream):
        self._content = stream

    @classmethod
    def read(cls, model, attribute_name):
        persisted_file_path = model.mapping.get(attribute_name)
        if not persisted_file_path:
            return super().read(model, attribute_name)

        extension = os.path.splitext(persisted_file_path)[1].lstrip(".")
        file_instance = cls(extension=extension)
        file_instance.associate(model, attribute_name)
        return file_instance

    def _is_new_model(self):
        return self.model is not None and self.attribute_name in self.model.mapping

    def relative_path(self):
        if not self._is_new_model():
            return super().relative_path()

        return self.model.mapping[self.attribute_name]

    def absolute_path(self):
        if not self._is_new_model():
            return super().absolute_path()

        return os.path.join(FILES_ROOT_DIR, self.relative_path())

    def write(self):
        super().write()
        self.model.mapping[self.attribute_name] = self.relative_path()

    def delete(self):
        if not self._is_new_model():
            super().delete()
        else:
            FilesGarbageCollector.instance().add(
                self.model.mapping[self.attribute_name]
            )
            del self.model.mapping[self.attribute_name]

    def __repr__(self):
        return "<FileObjectV2 model='{}' attribute_name='{}'>".format(
            self.model.__class__.__name__, self.attribute_name
        )


FileObject = FileObjectV2


class FileAttribute(BaseAttribute):
    """
    This attribute stores a file on disk.
    """

    def __init__(self, file_object_class=FileObject, **kwargs):
        self.FileObjectClass = file_object_class
        kwargs.setdefault("read_memoization", True)
        super().__init__(**kwargs)

    def read(self, parent):
        # TODO: Implement deserialize instead of read
        return self.FileObjectClass.read(parent, self.key(parent))

    def write(self, parent, value):
        # TODO: Implement serialize instead of write

        if not value:
            return value

        if isinstance(value, dict):
            value = self.FileObjectClass(**value)

        elif isinstance(value.model, BaseModel) and not BaseModel.__eq__(
            value.model, parent
        ):
            value = self.FileObjectClass(stream=value.stream, extension=value.extension)
        value.associate(parent, self.key(parent))

        return value

    def save(self, parent):
        _file = getattr(parent, self.key(parent))
        if _file:
            _file.write()
        else:
            _file = self.FileObjectClass.read(parent, self.key(parent))
            if _file:
                _file.delete()

    def delete(self, parent):
        _file = getattr(parent, self.key(parent))
        if not _file:
            _file = self.FileObjectClass()
            _file.associate(parent, self.key(parent))
        _file.delete()


class FilesGarbageCollector(object):
    DEFAULT_TABLE = "__sheraf_collected_files_to_remove"
    _instance = None

    @classmethod
    def instance(cls):
        if not cls._instance:
            cls._instance = cls()
        return cls._instance

    def __init__(self, table=DEFAULT_TABLE):
        self.table = table

    def _collected_paths(self):
        root = sheraf.Database.current_connection().root()
        if self.table not in root:
            root[self.table] = sheraf.types.Set()
        return root[self.table]

    @staticmethod
    def _relpath(path):
        if path.startswith(FILES_ROOT_DIR):
            return os.path.relpath(path, FILES_ROOT_DIR)
        return path

    def add(self, path):
        self._collected_paths().add(self._relpath(path))

    def remove(self, path):
        self._collected_paths().remove(self._relpath(path))

    def __contains__(self, path):
        return self._relpath(path) in self._collected_paths()

    def __bool__(self):
        return bool(self._collected_paths())

    def clear(self):
        for path in self._collected_paths():
            try:
                os.remove(os.path.join(FILES_ROOT_DIR, path))
            except OSError:
                pass
        self._collected_paths().clear()



import sheraf.types
import warnings


class BaseModelMetaclass(type):
    """
        Internal metaclass.
        Contains the mapping of attribute names with their corresponding data (of type :class:`~sheraf.attributes.BaseAttribute`)
    """

    def __new__(cls, name, bases, attrs):
        klass = super().__new__(cls, name, bases, attrs)
        klass.attributes = {}

        for name, attr in attrs.items():
            if not isinstance(attr, sheraf.attributes.base.BaseAttribute):
                continue

            try:
                attr.set_default_key(klass.attribute_id(name, attr))
            except NotImplementedError:
                continue

            klass.attributes[name] = attr

        for _base in bases:
            base_attributes = {}
            base_attributes.update(_base.__dict__.get("attributes", {}))
            base_attributes.update(_base.__dict__)
            for name, attr in base_attributes.items():
                if not isinstance(attr, sheraf.attributes.base.BaseAttribute):
                    continue

                if name in klass.attributes:
                    continue

                try:
                    attr.set_default_key(klass.attribute_id(name, attr))
                except NotImplementedError:
                    continue

                klass.attributes[name] = attr

        return klass


class BaseModel(object, metaclass=BaseModelMetaclass):
    """
    :class:`~sheraf.models.base.BaseModel` is the base class for every other model classes.
    This is where the attribute reading and writing are handled.

    Models can be used as dictionaries:

    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.SimpleAttribute()
    ...
    >>> with sheraf.connection(): # doctest: +SKIP
    ...     dict(Cowboy.create(name="George Abitbol"))
    {'name': 'George Abitbol', '_creation': ...}
    """

    attributes = {}
    mapping = None
    default_mapping = sheraf.types.SmallDict

    @classmethod
    def create(cls, default=None, *args, **kwargs):
        """Create a model instance.

        :param default: The data structure that will be used to store the model state.
        :param \\*\\*kwargs: Any model attribute can be initialized with the matching keyword.

        :return: The newly created model.

        >>> class Cowboy(sheraf.Model):
        ...    table = "cowboy"
        ...    name = sheraf.SimpleAttribute(default="John Doe")
        ...
        >>> with sheraf.connection(commit=True):
        ...    cowboy = Cowboy.create()
        ...    assert "John Doe" == cowboy.name
        ...
        ...    cowboy = Cowboy.create(name="George Abitbol")
        ...    assert "George Abitbol" == cowboy.name
        ...
        ...    Cowboy.create(this_attribute_does_not_exist="something") #  raises a TypeError
        Traceback (most recent call last):
            ...
        TypeError: TypeError: create() got an unexpected keyword argument 'this_attribute_does_not_exist'

        The function can also create sub-models recursively:

        >>> class Horse(sheraf.InlineModel):
        ...     name = sheraf.SimpleAttribute()
        ...
        >>> class Cowboy(sheraf.Model):
        ...     table = "cowboy"
        ...     name = sheraf.SimpleAttribute(default="John Doe")
        ...     horse = sheraf.InlineModelAttribute(Horse)
        ...
        >>> with sheraf.connection():
        ...     cowboy = Cowboy.create(name="George Abitbol", horse={"name": "Jolly Jumper"})
        ...     cowboy.horse.name
        'Jolly Jumper'
        """
        mapping = (default or cls.default_mapping)()
        model = cls._decorate(mapping)

        for attribute, value in kwargs.items():
            if attribute not in model.attributes:
                raise TypeError(
                    "TypeError: create() got an unexpected keyword argument '{}'".format(
                        attribute
                    )
                )
            model.__setattr__(attribute, value)

        for name, attribute in model.attributes.items():
            if not attribute.lazy and name not in kwargs:
                model.__setattr__(name, attribute.create(model))

        return model

    @classmethod
    def _decorate(cls, mapping):
        model = cls()
        model.mapping = mapping
        return model

    @classmethod
    def attribute_id(cls, name, attribute):
        raise NotImplementedError

    @property
    def _persistent(self):
        warnings.warn(
            "'_persistent' attribute has been renamed 'mapping'. "
            "The use of '_persistent' will stop working with sheraf 0.2",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.mapping

    def __setattr__(self, name, value):
        if name not in self.attributes:
            super().__setattr__(name, value)
            return

        value = self.attributes[name].write(self, value)
        if self.attributes[name].write_memoization:
            super().__setattr__(name, value)

    def __delattr__(self, name):
        if name in self.attributes:
            try:
                self.attributes[name].delattr(self)
                super().__delattr__(name)
            except (KeyError, AttributeError):
                return
        else:
            super().__delattr__(name)

    def __getattribute__(self, name):
        # TODO: Find a way to check that self.name exists or not without
        # using try/except AttributeError.
        # Because if a @property of this class, called `name` also raises
        # an AttributeError, actually we cannot say from where the AttributeError
        # was emitted.

        try:
            attribute = super().__getattribute__(name)
            if not isinstance(attribute, sheraf.attributes.base.BaseAttribute):
                return attribute

        except AttributeError as exc:
            if "object has no attribute '{}'".format(name) not in str(exc):
                raise

        try:
            attribute = self.attributes[name]
        except KeyError:
            raise AttributeError(name)

        value = attribute.read(self)

        if self.attributes[name].read_memoization:
            super().__setattr__(name, value)

        return value

    def copy(self, **kwargs):
        """
        :param \*\*kwargs: Keywords arguments will be passed to
                           :func:`~sheraf.models.BaseModel.create` and thus
                           wont be copied.

        :return: a copy of this instance.
        """
        copy = self.__class__.create(**kwargs)
        for name, attr in self.attributes.items():
            if attr.is_created(self) and name not in kwargs:
                setattr(copy, name, getattr(self, name))
        return copy

    def keys(self):
        """
        :return: The model attribute names.
        """
        return self.attributes.keys()

    def items(self):
        return (
            (key, self.attributes[key].read(self)) for key in self.attributes.keys()
        )

    def update(self, **kwargs):
        """Takes an arbitrary number of keywords arguments, and updates the
        model attributes matching the arguments.

        This functions recursively calls :func:`sheraf.attributes.base.BaseAttribute.edit` with `addition` and `edition` to `True`.

        >>> class Horse(sheraf.InlineModel):
        ...     name = sheraf.SimpleAttribute()
        ...
        >>> class Cowboy(sheraf.Model):
        ...     table = "people"
        ...     name = sheraf.SimpleAttribute()
        ...     horse = sheraf.InlineModelAttribute(Horse)
        ...
        >>> with sheraf.connection(commit=True):
        ...     george = Cowboy.create(name="George", horse={"name": "Centaurus"})
        ...     george.update(name="*incognito*", horse={"name": "Jolly Jumper"})
        ...     george.name
        '*incognito*'
        >>> with sheraf.connection():
        ...     george.horse.name
        'Jolly Jumper'

        Note that sub-models are also edited.
        """
        self.edit(
            kwargs, addition=True, edition=True, deletion=False, replacement=False
        )

    def assign(self, **kwargs):
        """Takes an arbitrary number of keywords arguments, and updates the
        model attributes matching the arguments.

        This functions recursively calls :func:`sheraf.attributes.base.BaseAttribute.edit` with `addition`, `edition` and `deletion` to `True`.

        >>> class Arm(sheraf.InlineModel):
        ...     name = sheraf.SimpleAttribute()
        ...
        >>> class Cowboy(sheraf.Model):
        ...     table = "people"
        ...     name = sheraf.SimpleAttribute()
        ...     arms = sheraf.SmallListAttribute(sheraf.InlineModelAttribute(Arm))
        ...
        >>> with sheraf.connection(commit=True):
        ...     george = Cowboy.create(name="Three arms cowboy", arms=[
        ...         {"name": "Arm 1"}, {"name": "Arm 2"}, {"name": "Arm 3"},
        ...     ])
        ...     len(george.arms)
        3
        >>> with sheraf.connection(commit=True):
        ...     george.assign(name="George Abitbol", arms=[
        ...         {"name": "Superarm 1"}, {"name": "Superarm 2"},
        ...     ])
        ...     len(george.arms)
        2
        >>> with sheraf.connection():
        ...     george.arms[0].name
        'Superarm 1'

        George passed from 3 arms to only 2 because *assign* does remove sub models. If we had called :func:`~sheraf.models.base.BaseModel.update` instead,
        George would have his two first arms be renamed *superarms* but, the third one would not have been removed.
        """
        self.edit(kwargs, addition=True, edition=True, deletion=True, replacement=False)

    def edit(
        self, value, addition=True, edition=True, deletion=False, replacement=False
    ):
        """Take a dictionary and a set of options, and try to applies the
        dictionary values to the model structure.

        :param value: The dictionary containing the values. The dictionary elements that
                      do not match the model attributes will be ignored.
        :param addition: If *True*, elements present in *value* and absent from the model attributes will be added.
        :param edition: If *True*, elements present in both *value* and the model will be updated.
        :param deletion: If *True*, elements present in the model and absent from *value* will be deleted.
        :param replacement: Like *edition*, but create a new element instead of updating one.
        """
        try:
            for attr, new_value in value.items():
                old_value = self.attributes[attr].read(self)
                updated = self.attributes[attr].update(
                    old_value, new_value, addition, edition, deletion, replacement
                )
                self.__setattr__(attr, updated)
        except KeyError:
            raise TypeError(
                "TypeError: edit() got an unexpected keyword argument '{}'".format(attr)
            )
        return self

    def save(self):
        for attr in self.attributes.values():
            attr.save(self)
        return self

    def reset(self, attribute):
        """
        Resets an attribute to its default value.
        """
        self.__setattr__(attribute, self.attributes[attribute].create(self))

    def __ne__(self, other):
        return not self == other

    def __eq__(self, other):
        return (
            hasattr(self, "mapping")
            and hasattr(other, "mapping")
            and self.mapping == other.mapping
        )

    def __getitem__(self, key):
        return self.attributes[key].read(self)

    def __setitem__(self, key, value):
        value = self.attributes[key].write(self, value)
        super().__setattr__(key, value)

    def __contains__(self, key):
        return key in self.attributes

    def __repr__(self):
        return "<{}>".format(self.__class__.__name__)



import sheraf.exceptions
import sheraf.models
import warnings


class Index(sheraf.models.Model):
    """usage :

    class UserIndex(sheraf.Index):
        table = 'users_index'
        index_dict = sheraf.LargeDictAttribute(sheraf.ModelAttribute(User))
        key = lambda index, model: model.email
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "sheraf.indexes.Index support will stop with sheraf 0.2."
            "If you still want to use them, please copy the code in your projects.",
            DeprecationWarning,
            stacklevel=8,
        )
        super().__init__(*args, **kwargs)

    @classmethod
    def add(cls, model):
        return cls._get_instance()._add(model)

    @classmethod
    def get(cls, key):
        return cls._get_instance()._get(key)

    @classmethod
    def delete(cls, model):
        return cls._get_instance()._delete(model)

    @classmethod
    def delete_key(cls, key):
        return cls._get_instance()._delete_key(key)

    @classmethod
    def _get_instance(cls):
        try:
            return next(cls.all())
        except StopIteration:
            return cls.create()

    def _add(self, model):
        self.index_dict[self.key(model)] = model
        return model

    def _get(self, key):
        try:
            return self.index_dict[key]
        except KeyError:
            model = (
                self.attributes["index_dict"].attribute._model
                if hasattr(self.attributes["index_dict"], "attribute")
                else self.attributes["index_dict"].model
            )
            raise sheraf.exceptions.IndexObjectNotFoundException(self, key, model)

    def _delete(self, model):
        self._delete_key(self.key(model))

    def _delete_key(self, key):
        del self.index_dict[key]



import functools
import logging
import random
import time

import ZODB.POSException

import sheraf

ATTEMPTS = 5
logger = logging.getLogger(__name__)


def attempt(
    function,
    attempts=ATTEMPTS,
    commit=lambda: True,
    log_exceptions=True,
    args=None,
    kwargs=None,
    also_except=None,
    on_failure=lambda nb_attempt: time.sleep(random.uniform(0, nb_attempt)),
):
    """This method attemps to execute a function several times until no
    ConflictError is encountered.

    :param function: The function to execute.
    :param args: The positionnal arguments to pass to the function.
    :param kwargs: The keyword arguments to pass to the function.
    :param attempts: The number of attempts.
    :param commit: An executable returning a boolean indicating wether to
        commit after executing `function`.
    :param log_exceptions: A boolean indicating wether to log exceptions.
    :param also_except: A tuple containing exception classes. If one of those
        exceptions is catched during the attempt, a new execution will be
        attempted.
    :param on_failure: Callback called after each failed attempt except the
        last one. By default it waits a random time between 0 and `attempts`.
        The callback takes one argument ``nb_attempt``.

    :return: The return value of `function`.

    Example:

    >>> with sheraf.connection():
    ...     try:
    ...         sheraf.attempt(do_amazing_stuff, attempts=5, args=("DANGER",))
    ...     except ZODB.POSException.ConflictError:
    ...         print("Impossible to commit do_amazing_stuff after 5 attempts.")
    ...
    >>> with sheraf.connection():
    ...     try:
    ...         sheraf.attempt(lambda: 1/0, also_except=(ZeroDivisionError,))
    ...     except ZeroDivisionError:
    ...         print("This has very few chances to be successful ¯\\\\_(ツ)_/¯")
    This has very few chances to be successful ¯\\_(ツ)_/¯
    """

    connection = sheraf.Database.current_connection()
    if not connection:
        raise sheraf.exceptions.NotConnectedException()

    args = args if args is not None else ()
    kwargs = kwargs if kwargs is not None else {}
    if also_except:
        also_except = also_except if isinstance(also_except, tuple) else (also_except,)
    exception_classes = (ZODB.POSException.ConflictError,) + (also_except or ())

    _exc = Exception
    for x in range(attempts):
        start_time = time.time()
        start_commit_time = None
        try:
            connection.transaction_manager.begin()
            _response = function(*args, **kwargs)
            if commit():
                start_commit_time = time.time()
                connection.transaction_manager.commit()
            else:
                connection.transaction_manager.abort()
            return _response

        except exception_classes as exc:
            # Not so easy to change the message of an exception. Just adding an 'extra_info' field
            # https://stackoverflow.com/questions/6062576/adding-information-to-an-exception
            # https://stackoverflow.com/questions/9157210/how-do-i-raise-the-same-exception-with-a-custom-message-in-python/9157277
            now = time.time()
            execution_time = now - start_time

            if start_commit_time:
                commit_time = now - start_commit_time
                exc.extra_info = "Execution n°{} took {:.5f}s of which {:5f}s for the main function and {:5f}s to commit\n".format(
                    x + 1, execution_time, execution_time - commit_time, commit_time
                )
            else:
                exc.extra_info = "Execution n°{} took {:.5f}s\n".format(
                    x + 1, execution_time
                )

            if log_exceptions:
                extra = {
                    "attempt": x + 1,
                    "exception": exc,
                    "total_time": execution_time,
                    "stack": True,  # Sentry parameter: show full stack
                }
                if start_commit_time:
                    extra["commit_time"] = commit_time
                logger.warning(exc.extra_info, exc_info=True, extra=extra)

            connection.transaction_manager.abort()

            if x != attempts - 1:
                if on_failure:
                    on_failure(x)

            if x > 0:
                exc.extra_info += _exc.extra_info

            _exc = exc

        except Exception:
            connection.transaction_manager.abort()
            raise
    raise _exc


def commit(f):
    """
    Wrapper shortcut for :func:`~sheraf.transactions.attempt`.

    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     dead = sheraf.BooleanAttribute(default=False)
    ...
    >>> @sheraf.commit
    ... def fight(winner, loser):
    ...      loser.dead = True
    ...
    >>> with sheraf.connection(commit=True):
    ...     winner = Cowboy.create()
    ...     loser = Cowboy.create()
    ...
    >>> with sheraf.connection():
    ...     fight(winner, loser)
    ...
    >>> with sheraf.connection():
    ...     Cowboy.read(loser.id).dead
    True
    """

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return attempt(f, args=args, kwargs=kwargs)

    return wrapper



import ZODB

import sheraf
from sheraf.batches.utils import discover_models

try:  # pragma: no cover
    import colored

    HAS_COLORED = True
except ImportError:  # pragma: no cover
    HAS_COLORED = False

try:  # pragma: no cover
    from tqdm import tqdm

    HAS_TQDM = True
except ImportError:  # pragma: no cover
    HAS_TQDM = False


def check_conflict_resolution():
    """A simple utility functions that checks that sheraf objects conflict
    resolution is active.

    :return: ``True`` if sheraf objects conflicts can be solved, else ``False``.

    >>> from sheraf.batches.checks import check_conflict_resolution
    >>> with sheraf.connection():
    ...    assert check_conflict_resolution()
    """
    table_name = "__conflict_resolution_test_model__"
    nestable = sheraf.Database.get().nestable
    sheraf.Database.get().nestable = True

    class TestModel(sheraf.Model):
        table = table_name
        counter = sheraf.CounterAttribute(default=0)

    with sheraf.connection(commit=True):
        m = TestModel.create()

    try:
        with sheraf.connection(commit=True):
            TestModel.read(m.identifier).counter.increment(1)

            with sheraf.connection(commit=True):
                TestModel.read(m.identifier).counter.increment(15)

    except ZODB.POSException.ConflictError:  # pragma: no cover
        return False

    else:
        return True

    finally:
        with sheraf.connection(commit=True) as conn:
            del conn.root()[table_name]

        sheraf.Database.get().nestable = nestable


# Set the list of check functions to be run
INSTANCE_CHECK_FUNCS = {}
ATTRIBUTE_CHECK_FUNCS = {}
MODEL_CHECK_FUNCS = {}
OTHER_CHECK_FUNCS = {"check_conflict_resolution": check_conflict_resolution}


def check_health(
    *args,
    model_checks=None,
    instance_checks=None,
    attribute_checks=None,
    other_checks=None
):
    """
    Takes some modules in parameters.
    :param model_checks: If None, check also model consistency (see constant model_check_funcs)
    :param instance_checks: If None, check all instance consistency rules. Else, give the list of wanted
    checking rules (see constant instance_check_funcs)
    :param attribute_checks: If None, check all attribute consistency rules. Else, give the list of wanted
    checking rules (see constant instance_check_funcs)
    The function will discover models in the modules, analyze every
    model instance, and return a health report in JSON. Depending on the
    result, you may need to apply some migrations.
    """
    models = discover_models(*args)
    health_report = {}

    instance_checks = instance_checks or []
    attribute_checks = attribute_checks or []
    model_checks = model_checks or []
    other_checks = other_checks or []
    if not instance_checks and not attribute_checks and not model_checks:
        instance_checks = INSTANCE_CHECK_FUNCS.keys()
        attribute_checks = ATTRIBUTE_CHECK_FUNCS.keys()
        model_checks = MODEL_CHECK_FUNCS.keys()
        other_checks = OTHER_CHECK_FUNCS.keys()

    for check_key in other_checks:
        health_report[check_key] = OTHER_CHECK_FUNCS[check_key]()

    for model_path, model in models:

        for check_key in instance_checks:
            health_report.setdefault(
                INSTANCE_CHECK_FUNCS[check_key].__name__, {}
            ).setdefault(model_path, {"ok": 0, "ko": 0})

        for check_key in model_checks:
            model_check_func = MODEL_CHECK_FUNCS[check_key]
            health_report.setdefault(model_check_func.__name__, {})[
                model_path
            ] = model_check_func(model)

        # Iterate on instances
        if HAS_TQDM:  # pragma: no cover
            iterator = tqdm(model.all(), total=model.count(), desc=model.__name__)
        else:
            iterator = model.all()
        for m in iterator:
            for check_key in instance_checks:
                check_func = INSTANCE_CHECK_FUNCS[check_key]
                model_instance_result = health_report[check_func.__name__][model_path]
                if check_func(m):
                    model_instance_result["ok"] += 1
                else:
                    model_instance_result["ko"] += 1

            for check_key in attribute_checks:
                check_func = ATTRIBUTE_CHECK_FUNCS[check_key]
                health_report.setdefault(check_func.__name__, {})
                for attribute_name, bool_value in check_func(m).items():
                    attribute_result = (
                        health_report[check_func.__name__]
                        .setdefault(model_path, {})
                        .setdefault(attribute_name, {"ok": 0, "ko": 0})
                    )
                    if bool_value:
                        attribute_result["ok"] += 1
                    else:
                        attribute_result["ko"] += 1
    return health_report


def _print(string, padding, color):
    if not HAS_COLORED:
        return str(string).ljust(padding, "_")

    if padding:  # pragma: no cover
        return (
            colored.stylize(string, colored.fg("blue"))
            + (padding - len(str(string))) * "_"
        )
    return colored.stylize(string, colored.fg("blue"))  # pragma: no cover


def print_neutral(string, padding=0):
    return _print(string, padding, "blue")


def print_success(string, padding=0):
    return _print(string, padding, "green")


def print_failure(string, padding=0):
    return _print(string, padding, "red")


def _print_check_other_health_result(check_reason, health_table):
    print(
        "- {:_<86} ".format(print_neutral(check_reason))
        + " "
        + (print_success("OK") if health_table[check_reason] else print_failure("KO"))
    )


def _print_check_model_health_result(check_reason, health_table):

    print(80 * "=" + "\n" + print_neutral(check_reason) + 39 * " " + """OK       KO""")
    table_key = "check_model_" + check_reason
    for model_path, attributes in health_table[table_key].items():
        print(
            "- {:_<52} TOTAL: ".format(model_path)
            + print_success(sum(values["ok"] for values in attributes.values()), 8)
            + " "
            + print_failure(sum(values["ko"] for values in attributes.values()), 8)
        )

        for attribute_name, values in attributes.items():
            print(
                "  - {:_<57} ".format(attribute_name)
                + print_success(values["ok"], 8)
                + " "
                + print_failure(values["ko"], 8)
            )

    if not health_table[table_key]:
        print("  No model to visit.")


def _print_check_instance_health_result(check_reason, health_table):
    """
    :param check_reason: one among model_checks keys
    :param health_table: result of a check function
    """
    print(80 * "=" + "\n" + print_neutral(check_reason) + 39 * " " + """OK       KO""")
    for model_path, values in health_table["check_instance_" + check_reason].items():
        print(
            "- {:_<59} ".format(model_path)
            + print_success(values["ok"], 8)
            + " "
            + print_failure(values["ko"], 8)
        )


def _print_check_attribute_health_result(check_reason, health_table):
    print(80 * "=" + "\n{:<62}".format(print_neutral(check_reason)) + """OK       KO""")
    table_key = "check_attributes_" + check_reason
    for model_path, attributes in health_table[table_key].items():
        print(
            "- {:_<52} TOTAL: ".format(model_path)
            + print_success(sum(values["ok"] for values in attributes.values()), 8)
            + " "
            + print_failure(sum(values["ko"] for values in attributes.values()), 8)
        )

        for attribute_name, values in attributes.items():
            print(
                "  - {:_<57} ".format(attribute_name)
                + print_success(values["ok"], 8)
                + " "
                + print_failure(values["ko"], 8)
            )

    if not health_table[table_key]:
        print("  No model to visit.")


def print_health(
    *args,
    model_checks=None,
    instance_checks=None,
    attribute_checks=None,
    other_checks=None
):
    """Takes some modules in parameters (e.g. "american.class.cowboy_module").

    The function will discover models in the modules, analyze every model instance, and return
    a health report in a human readable format. Depending on the result, you may need to apply some migrations.

    This function does not edit any data and is safe to be executed in a production shell.
    """
    print(
        "             _                     __        _               _\n"
        "=========== | | ================= / _| ==== | | =========== | | ===============\n"
        "         ___| |__   ___ _ __ __ _| |_    ___| |__   ___  ___| | _____\n"
        "        / __| '_ \\ / _ \\ '__/ _` |  _|  / __| '_ \\ / _ \\/ __| |/ / __|\n"
        "        \\__ \\ | | |  __/ | | (_| | |   | (__| | | |  __/ (__|   <\\__ \\\n"
        "        |___/_| |_|\\___|_|  \\__,_|_|    \\___|_| |_|\\___|\\___|_|\\_\\___/\n"
        "==============================================================================="
    )
    print("Analyzing your models, this operation can be very long...")

    instance_checks = instance_checks or []
    attribute_checks = attribute_checks or []
    model_checks = model_checks or []
    other_checks = other_checks or []
    if not instance_checks and not attribute_checks and not model_checks:
        instance_checks = INSTANCE_CHECK_FUNCS.keys()
        attribute_checks = ATTRIBUTE_CHECK_FUNCS.keys()
        model_checks = MODEL_CHECK_FUNCS.keys()
        other_checks = OTHER_CHECK_FUNCS.keys()

    health = check_health(
        *args,
        model_checks=model_checks,
        instance_checks=instance_checks,
        attribute_checks=attribute_checks,
        other_checks=other_checks,
    )

    for other_check_type in other_checks:
        _print_check_other_health_result(other_check_type, health)

    for model_check_type in model_checks:
        _print_check_model_health_result(model_check_type, health)

    for instance_check_type in instance_checks:
        _print_check_instance_health_result(instance_check_type, health)

    for attribute_check_type in attribute_checks:
        _print_check_attribute_health_result(attribute_check_type, health)
